@@ -24,7 +24,7 @@ int main( void)
   /* For every font in the database, generate a glyph for the string
      "Test" at 25 bp. Use Kerning. Then dump an ASCII representation
      of the glyph to stdout */
-  for( i=0; i<T1_Get_no_fonts(); i++){
+  for( i=0; i<T1_GetNoFonts(); i++){
     printf("FontID=%d, Font=%s\n", i, T1_GetFontFilePath(i));
     printf("FontID=%d, Metrics=%s\n", i, T1_GetAfmFilePath(i));
     glyph=T1_SetString( i, "Test", 0, 0, T1_KERNING, 25.0, NULL);
