@@ -1,11 +1,11 @@
 /*--------------------------------------------------------------------------
   ----- File:        t1load.c 
-  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
-  ----- Date:        04/18/1998
+  ----- Author:      Rainer Menzner (Rainer.Menzner@web.de)
+  ----- Date:        2002-11-27
   ----- Description: This file is part of the t1-library. It contains
                      functions for loading fonts  and for managing size
 		     dependent data.
-  ----- Copyright:   t1lib is copyrighted (c) Rainer Menzner, 1996-1998. 
+  ----- Copyright:   t1lib is copyrighted (c) Rainer Menzner, 1996-2002. 
                      As of version 0.5, t1lib is distributed under the
 		     GNU General Public Library Lincense. The
 		     conditions can be found in the files LICENSE and
@@ -28,14 +28,19 @@
   
 #define T1LOAD_C
 
-/* #define REALLOC_VM */
 #define ANSI_REALLOC_VM  
 
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <unistd.h>
+#if defined(_MSC_VER)
+# include <io.h>
+# include <sys/types.h>
+# include <sys/stat.h>
+#else
+# include <unistd.h>
+#endif
 #include <stdlib.h>
 #include <math.h>
 #include <string.h>
@@ -55,15 +60,15 @@
 #include "t1extern.h"
 #include "t1load.h"
 #include "t1env.h"
+#include "t1set.h"
 #include "t1base.h"
 #include "t1finfo.h"
 #include "t1afmtool.h"
 
 
-extern char *vm_base; /* from fontfcn.c in initfont()! */
-extern char *vm_used; /* from fontfcn.c in fontfcnA()! */
 
-int fontfcnA( char *env, int *mode, psfont *Font_Ptr);
+extern psobj *StdEncArrayP;       /* For checking of a fonts encoding */
+extern char not_def[];            /* for checking the ".notdef"-string */
 
 
 
@@ -72,7 +77,7 @@ int fontfcnA( char *env, int *mode, psfont *Font_Ptr);
 
 int T1_LoadFont( int FontID)
 {
-  int i, j, k;
+  int i, j, k, l, m;
   char *FileName, *FileNamePath;
   int mode;  /* This is used by the type1-library for error reporting */   
   char *charname;
@@ -80,28 +85,28 @@ int T1_LoadFont( int FontID)
   /* The following vars are used for reallocation of VM */
   long tmp_size;
   float ascender;
-  unsigned long shift;
 #ifdef ANSI_REALLOC_VM
+  unsigned long shift;
   unsigned long ldummy;
-#endif
   char *tmp_ptr;
+#endif
 
-  extern psobj *StdEncArrayP;       /* For checking of a fonts encoding */
-  extern psobj *ISOLatin1EncArrayP;
+  struct region *area;
+  struct XYspace *S;    
 
   /* These are for constructing the kerning lookup table: */
   PairKernData *pkd;
-  KERN_ENTRY *kern_tbl;
+  METRICS_ENTRY *kern_tbl;
   int char1, char2;
   
   
-  if (CheckForInit()){
+  if (T1_CheckForInit()){
     T1_errno=T1ERR_OP_NOT_PERMITTED;
     return(-1);
   }
   
 
-  i=CheckForFontID(FontID);
+  i=T1_CheckForFontID(FontID);
   if (i==1)
     return(0);      /* Font already loaded */
   if (i==-1){
@@ -111,27 +116,23 @@ int T1_LoadFont( int FontID)
   
   /* Allocate memory for ps_font structure: */
   if ((pFontBase->pFontArray[FontID].pType1Data=(psfont *)malloc(sizeof(psfont)))==NULL){
-    sprintf( err_warn_msg_buf,
-	     "Failed to allocate memory for psfont-struct (FontID=%d)", FontID);
-    print_msg("T1_LoadFont()", err_warn_msg_buf);
+    T1_PrintLog( "T1_LoadFont()", "Failed to allocate memory for psfont-struct (FontID=%d)",
+		 T1LOG_ERROR, FontID);
     T1_errno=T1ERR_ALLOC_MEM;
     return(-1);
   }
 
   /* Check for valid filename */
   if ((FileName=T1_GetFontFileName(FontID))==NULL){
-    sprintf( err_warn_msg_buf, "No font file name for font %d",
-	     FontID);
-    print_msg( "T1_LoadFont()", err_warn_msg_buf);
+    T1_PrintLog( "T1_LoadFont()", "No font file name for font %d", T1LOG_ERROR, FontID);
     return(-1);
   }
   
   /* Fetch the full path of type1 font file */
-  if ((FileNamePath=Env_GetCompletePath( FileName,
+  if ((FileNamePath=intT1_Env_GetCompletePath( FileName,
 					 T1_PFAB_ptr))==NULL){
-    sprintf( err_warn_msg_buf, "Couldn't locate font file for font %d in %s",
-	     FontID, T1_PFAB_ptr);
-    print_msg( "T1_LoadFont()", err_warn_msg_buf);
+    T1_PrintLog( "T1_LoadFont()", "Couldn't locate font file for font %d in %s",
+		 T1LOG_ERROR, FontID, T1_GetFileSearchPath(T1_PFAB_PATH));
     T1_errno=T1ERR_FILE_OPEN_ERR;
     return(-1);
   }
@@ -139,10 +140,8 @@ int T1_LoadFont( int FontID)
   /* And load all PostScript information into memory */
   if (fontfcnA( FileNamePath, &mode,
 		pFontBase->pFontArray[FontID].pType1Data) == FALSE){
-    sprintf( err_warn_msg_buf,
-	     "Loading font with ID = %d failed! (mode = %d)",
-	    FontID, mode);
-    print_msg( "T1_LoadFont()", err_warn_msg_buf);
+    T1_PrintLog( "T1_LoadFont()", "Loading font with ID = %d failed! (mode = %d)",
+		 T1LOG_ERROR, FontID, mode);
     free(FileNamePath);
     pFontBase->pFontArray[FontID].pType1Data=NULL;
     T1_errno=mode;
@@ -154,104 +153,6 @@ int T1_LoadFont( int FontID)
   /* Store the base address of virtual memory and realloc in order not
      to waste too much memory: */
   pFontBase->pFontArray[FontID].vm_base=vm_base; 
-#ifdef REALLOC_VM
-  /* Get size of VM, ... */
-  tmp_size=(long)vm_used - (long)vm_base; 
-  /* Now, realloc.. */  
-  tmp_ptr=(char *)realloc(vm_base,  tmp_size); 
-  
-  /* ... and shift all pointers refering to that area */
-  
-  shift= (long)tmp_ptr - (long)vm_base;
-
-  /* We start by shifting the topmost pointers: */
-  pFontBase->pFontArray[FontID].vm_base=tmp_ptr;
-  (unsigned char *)(pFontBase->pFontArray[FontID].pType1Data->vm_start) +=shift;
-  (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->CharStringsP +=shift;
-  (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->Private +=shift;
-  (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP +=shift;
-  (unsigned char *)(pFontBase->pFontArray[FontID].pType1Data->BluesP) +=shift;
-  
-  /* We now have to care for correcting all pointers which are in the VM
-     and refer to some place in the VM! Note: Instead of selecting the
-     appropriate pointer-elements of the union we simply shift the unspecified
-     pointer "valueP".
-     Note: The filename entry does not need to be modified since it does not
-     need to be shifted since it points to memory managed by t1lib.
-     */
-  /* FontInfo-dictionary: All name-pointers and the pointers to all array
-     types have to be shifted: */
-  i=pFontBase->pFontArray[FontID].pType1Data->fontInfoP[0].key.len;
-  for (j=1; j<=i; j++){
-    if ((pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.type==OBJ_ARRAY) ||
-	(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.type==OBJ_STRING) ||
-	(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.type==OBJ_NAME) ||
-	(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.type==OBJ_FILE)){
-      (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP +=shift;
-    }
-    /* The encoding needs special treatment: */
-    if (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.type==OBJ_ENCODING){
-      /* If a builtin encoding is used, it is sufficient to shift the pointer
-	 to the Encoding since the character-namestrings of builtin encodings
-	 are static and thus located on the heap.
-	 For font-specific encoding, character-namestrings reside in VM and
-	 thus each entry has to be shifted. 
-	 Caution: We still have to shift the builtin encoding-pointer, since
-	 they also point to are located in VM: */
-      (unsigned char *)StdEncArrayP +=shift;
-      (unsigned char *)ISOLatin1EncArrayP +=shift;
-      (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP +=shift;
-      if ((pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	   == StdEncArrayP) ||
-	  (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	   == ISOLatin1EncArrayP)){ /* Font uses builtin encoding */
-	;
-      } 
-      else{ /* Font-specific encoding */ 
-	for (k=0; k<256; k++){
-	  (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP += shift;
-	}
-      }
-    } /* end of encoding-handling */
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].key.data.valueP +=shift;
-  } /* fontinfo-dict done */
-
-  /* Private-dictionary: All name-pointers and the pointers to all array
-     types have to be shifted: */
-  i=pFontBase->pFontArray[FontID].pType1Data->Private[0].key.len;
-  for (j=1; j<=i; j++){
-    if ((pFontBase->pFontArray[FontID].pType1Data->Private[j].value.type==OBJ_ARRAY) ||
-	(pFontBase->pFontArray[FontID].pType1Data->Private[j].value.type==OBJ_STRING) ||
-	(pFontBase->pFontArray[FontID].pType1Data->Private[j].value.type==OBJ_NAME) ||
-	(pFontBase->pFontArray[FontID].pType1Data->Private[j].value.type==OBJ_FILE)){
-      (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->Private[j].value.data.valueP +=shift;
-    }
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->Private[j].key.data.valueP +=shift;
-  }
-  
-  /* BluesP: The entry "next" is the only pointer in blues_struct. Although it is
-     not used anywhere we should shift it for correctness reasons (in case its not
-     NULL)! */
-  if (pFontBase->pFontArray[FontID].pType1Data->BluesP->next != NULL)
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->BluesP->next +=shift;
-
-  /* The CharStrings-dictionary: Every namepointer and its corresponding
-     charstring has to be shifted: */
-  i=pFontBase->pFontArray[FontID].pType1Data->CharStringsP[0].key.len;
-  for (j=1; j<=i; j++){
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->CharStringsP[j].value.data.valueP +=shift;
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->CharStringsP[j].key.data.valueP +=shift;
-  }
-
-  /* The Subroutines have also to be reorganized: */
-  i=pFontBase->pFontArray[FontID].pType1Data->Subrs.len;
-  /* First, shift pointer to array-start and after that the pointers to
-     each command string: */
-  (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP +=shift;
-  for (j=1; j<=i; j++) {
-    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP[j].data.valueP +=shift;
-  }
-#endif
 #ifdef ANSI_REALLOC_VM
   /* We first get the size of pointers on the current system */
   /* Get size of VM, ... */
@@ -320,23 +221,21 @@ int T1_LoadFont( int FontID)
 	ldummy=(long)StdEncArrayP;
 	ldummy +=shift;
 	StdEncArrayP=(psobj *)ldummy;
-	ldummy=(long)ISOLatin1EncArrayP;
-	ldummy +=shift;
-	ISOLatin1EncArrayP=(psobj *)ldummy;
 	ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP;
 	ldummy +=shift;
 	pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP=(char *)ldummy;
-	if ((pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	     == StdEncArrayP) ||
-	    (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	     == ISOLatin1EncArrayP)){ /* Font uses builtin encoding */
+	if (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
+	    == StdEncArrayP){ /* Font uses builtin standard encoding */
 	  ;
 	} 
 	else{ /* Font-specific encoding */ 
 	  for (k=0; k<256; k++){
 	    ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP;
-	    ldummy +=shift;
-	    pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP=(struct ps_obj *)ldummy;
+	    /* The ".notdef" is also static and may not be shifted (Thanks, Derek ;) */
+	    if (ldummy != (unsigned long)not_def) {
+	      ldummy +=shift;
+	      pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP=(struct ps_obj *)ldummy;
+	    }
 	  }
 	}
       } /* end of encoding-handling */
@@ -390,7 +289,7 @@ int T1_LoadFont( int FontID)
     ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP;
     ldummy +=shift;
     pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP=(struct ps_obj *)ldummy;
-    for (j=1; j<=i; j++) {
+    for (j=0; j<i; j++) {
       ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP[j].data.valueP;
       ldummy +=shift;
       pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP[j].data.valueP=(char *)ldummy;
@@ -457,23 +356,21 @@ int T1_LoadFont( int FontID)
 	ldummy=(long)StdEncArrayP;
 	ldummy -=shift;
 	StdEncArrayP=(psobj *)ldummy;
-	ldummy=(long)ISOLatin1EncArrayP;
-	ldummy -=shift;
-	ISOLatin1EncArrayP=(psobj *)ldummy;
 	ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP;
 	ldummy -=shift;
 	pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.valueP=(char *)ldummy;
-	if ((pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	     == StdEncArrayP) ||
-	    (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
-	     == ISOLatin1EncArrayP)){ /* Font uses builtin encoding */
+	if (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP
+	     == StdEncArrayP){ /* Font uses builtin encoding */
 	  ;
 	} 
 	else{ /* Font-specific encoding */ 
 	  for (k=0; k<256; k++){
 	    ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP;
-	    ldummy -=shift;
-	    pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP=(struct ps_obj *)ldummy;
+	    /* The ".notdef" is also static and may not be shifted (Thanks, Derek ;) */
+	    if (ldummy != (unsigned long)not_def) {
+	      ldummy -=shift;
+	      pFontBase->pFontArray[FontID].pType1Data->fontInfoP[j].value.data.arrayP[k].data.arrayP=(struct ps_obj *)ldummy;
+	    }
 	  }
 	}
       } /* end of encoding-handling */
@@ -527,7 +424,7 @@ int T1_LoadFont( int FontID)
     ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP;
     ldummy -=shift;
     pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP=(struct ps_obj *)ldummy;
-    for (j=1; j<=i; j++) {
+    for (j=0; j<i; j++) {
       ldummy=(long)pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP[j].data.valueP;
       ldummy -=shift;
       pFontBase->pFontArray[FontID].pType1Data->Subrs.data.arrayP[j].data.valueP=(char *)ldummy;
@@ -576,51 +473,79 @@ int T1_LoadFont( int FontID)
   pFontBase->pFontArray[FontID].extend=1.0;
 
   
-  /* Now try to load afm-structures from corresponding .afm-file. */
-  if ((i=openFontMetricsFile( FontID))){
-    /* Try a fallback function: */
-    if ((i=openFontMetricsFileSloppy( FontID))){
-      sprintf( err_warn_msg_buf,
-	       "Alert: Error (%d) sloppy-processing afm-file for Font %d!",
-	       i ,FontID);
-      T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_STATISTIC);
-      if ((pFontBase->pFontArray[FontID].pAFMData=
-	   T1_GenerateAFMFallbackInfo(FontID))==NULL){
+  /* Now try to load afm-structures from corresponding .afm-file (if
+     not suppressed by the user). */
+  if ((pFontBase->t1lib_flags & T1_NO_AFM)!=0) {
+    pFontBase->pFontArray[FontID].pAFMData = NULL;
+    T1_PrintLog( "T1_LoadFont()",
+		 "Suppressing AFM data handling on user request",
+		 T1LOG_STATISTIC);
+  }
+  else {
+    if ((i=openFontMetricsFile( FontID, 0))){
+      /* Try a fallback, opening sloppy: */
+      if ((i=openFontMetricsFile( FontID, 1))) {
 	sprintf( err_warn_msg_buf,
-		 "Ultimately failed to generate metrics information Font %d!",
-		 FontID);
-	T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_WARNING);
-	T1_PrintLog( "T1_LoadFont()", "Segmentation fault might result",T1LOG_WARNING);
-      }
-      else{
-	pFontBase->pFontArray[FontID].info_flags |=AFM_SELFGEN_SUCCESS;
-	T1_PrintLog( "T1_LoadFont()",
-		     "Generating AFM-information from fontfile successful!",
-		     T1LOG_STATISTIC);
+		 "Alert: Error (%d) sloppy-processing afm-file for Font %d!",
+		 i ,FontID);
+	T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_STATISTIC);
+	if ((pFontBase->pFontArray[FontID].pAFMData=
+	     T1_GenerateAFMFallbackInfo(FontID))==NULL){
+	  sprintf( err_warn_msg_buf,
+		   "Ultimately failed to generate metrics information Font %d!",
+		   FontID);
+	  T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_WARNING);
+	}
+	else {
+	  pFontBase->pFontArray[FontID].info_flags |=AFM_SELFGEN_SUCCESS;
+	  T1_PrintLog( "T1_LoadFont()",
+		       "Generating AFM-information from fontfile successful!",
+		       T1LOG_STATISTIC);
+	}
       }
+      else {
+	pFontBase->pFontArray[FontID].info_flags |=AFM_SLOPPY_SUCCESS;
+	sprintf( err_warn_msg_buf,
+		 "Alert: Limited afm-information for Font %d",FontID);
+	T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_STATISTIC);
+      } 
     }
-    else{
-      pFontBase->pFontArray[FontID].info_flags |=AFM_SLOPPY_SUCCESS;
-      sprintf( err_warn_msg_buf,
-	       "Alert: Limited afm-information for Font %d",FontID);
-      T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_STATISTIC);
+    else {
+      pFontBase->pFontArray[FontID].info_flags |=AFM_SUCCESS;
     }
   }
-  else
-    pFontBase->pFontArray[FontID].info_flags |=AFM_SUCCESS;
   
-  /* Now, set Encodingvector entry to NULL, thus indicating that by default the
-     fonts internal encoding should be used
+  
+  /* Now, set Encodingvector entry to default if the font's
+     internal encoding is "StandardEncoding".
      */
-  pFontBase->pFontArray[FontID].pFontEnc=NULL;
+  if (pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP
+      == StdEncArrayP) {
+    pFontBase->pFontArray[FontID].info_flags |=USES_STANDARD_ENCODING;
+    pFontBase->pFontArray[FontID].pFontEnc=pFontBase->default_enc;
+    sprintf( err_warn_msg_buf,
+	     "Font %d reencoded to default",FontID);
+    T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_DEBUG);
+  }
+  else {
+    sprintf( err_warn_msg_buf,
+	     "Font %d not reencoded to default",FontID);
+    T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf, T1LOG_DEBUG);
+    pFontBase->pFontArray[FontID].pFontEnc = NULL;
+  }
 
   
   /* If AFM-Info available we try to speed up some things: */
-  if (pFontBase->pFontArray[FontID].pAFMData != NULL){
+  if (pFontBase->pFontArray[FontID].pAFMData != NULL) {
     /* We have to fill the array that maps the current encodings' indices to the
-       indices used in afm file */
+       indices used in afm file. The interpretation has been changed in
+       in t1lib-1.2. We now use positive values for indexing into the charmetrics
+       array and negative values for indexing into the composite character array.
+       an index of zero indicates that no metrics are defined for this character.
+       This may happen because (a) not all AFM-files define metrics for the .notdef
+       character, and (b) because font and AFM-file do not match. */
     if ((pFontBase->pFontArray[FontID].pEncMap=
-	 (int *)calloc(256,sizeof(int)))==NULL){
+	 (int *)calloc(256,sizeof(int)))==NULL) {
       sprintf( err_warn_msg_buf, "Error allocating memory for encoding map (FontID=%d)",
 	       FontID);
       T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf,
@@ -628,38 +553,72 @@ int T1_LoadFont( int FontID)
       T1_errno=T1ERR_ALLOC_MEM;
       return(-1);
     }
-    for (i=0; i<256; i++)
-      pFontBase->pFontArray[FontID].pEncMap[i]=-1;
-    for (i=0; i<256; i++){
+    for (i=0; i<256; i++) {
       charname=T1_GetCharName( FontID, i);
-      for ( j=0; j<pFontBase->pFontArray[FontID].pAFMData->numOfChars; j++){
+      /* in a first loop check for ordinary characters */
+      for ( j=0; j<pFontBase->pFontArray[FontID].pAFMData->numOfChars; j++) {
 	if (strcmp( charname,
-		    pFontBase->pFontArray[FontID].pAFMData->cmi[j].name)==0){
-	  pFontBase->pFontArray[FontID].pEncMap[i]=j;
+		    pFontBase->pFontArray[FontID].pAFMData->cmi[j].name)==0) {
+	  pFontBase->pFontArray[FontID].pEncMap[i]=j+1; /* index 0 is reserved! */
+	  continue;
+	}
+      }
+      /* if nothing has been found, check for composite characters */ 
+      for ( j=0; j<pFontBase->pFontArray[FontID].pAFMData->numOfComps; j++) {
+	if (strcmp( charname,
+		    pFontBase->pFontArray[FontID].pAFMData->ccd[j].ccName)==0) {
+	  pFontBase->pFontArray[FontID].pEncMap[i]=-(j+1); /* index 0 is reserved! */
+	  continue;
 	}
       }
     }
     
-    /* We now introduce a "flag map" which gives quickly tells what metric 
-       features a character provides. */
-    if ((pFontBase->pFontArray[FontID].pFlagMap= 
-	 (char *)calloc(256,sizeof(char)))==NULL){ 
-      sprintf( err_warn_msg_buf, "Error allocating memory for flag map (FontID=%d)",
-	       FontID);
-      T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf,
-		 T1LOG_WARNING);
-      T1_errno=T1ERR_ALLOC_MEM;
-      return(-1); 
-    } 
-    
+    /* For composite characters, we still have to compute the width and bbox */
+    for ( j=0; j<pFontBase->pFontArray[FontID].pAFMData->numOfComps; j++) {
+      /*and bounding box by ourselves. First, set up an identity charspace
+	matrix and then generate an edgelist for the composite character at
+	size 1000bp using no transformation and current encoding. Note: This
+	action is only required when loading a font at first time, but not
+	when reencoding a font. */
+      S=(struct XYspace *)IDENTITY;
+      S=(struct XYspace *)Permanent
+	(Transform(S, pFontBase->pFontArray[FontID].FontTransform[0],
+		   pFontBase->pFontArray[FontID].FontTransform[1],
+		   pFontBase->pFontArray[FontID].FontTransform[2],
+		   pFontBase->pFontArray[FontID].FontTransform[3]));
+      
+      area=fontfcnB_ByName( FontID, 0, S,
+			    pFontBase->pFontArray[FontID].pAFMData->ccd[j].ccName,
+			    &mode, pFontBase->pFontArray[FontID].pType1Data,
+			    DO_RASTER);
+      /* Store bounding box ... */
+      pFontBase->pFontArray[FontID].pAFMData->ccd[j].charBBox.llx=area->xmin;
+      pFontBase->pFontArray[FontID].pAFMData->ccd[j].charBBox.urx=area->xmax;
+      pFontBase->pFontArray[FontID].pAFMData->ccd[j].charBBox.lly=area->ymin;
+      pFontBase->pFontArray[FontID].pAFMData->ccd[j].charBBox.ury=area->ymax;
+      /* ... and character width. This should be the width of the base character
+	 of the composite! */
+      pFontBase->pFontArray[FontID].pAFMData->ccd[j].wx=NEARESTPEL(area->ending.x);
+      /* clean up. */
+      KillRegion (area);
+      if (S!=NULL) {
+	KillSpace (S);
+	S=NULL;
+      }
+    }
     /* We now create an encoding-specific kerning table which will speed up
        looking for kerning pairs! */
+    pFontBase->pFontArray[FontID].KernMapSize=0;
     /* First, get number of defined kerning pairs: */
     k=pFontBase->pFontArray[FontID].pAFMData->numOfPairs;
-    if (k>0){ /* i.e., there are any pairs */ 
+    if (k>0){ /* i.e., there are any pairs */
+      /* OK, it does not suffice to alloc numOfPairs METRICS_ENTRYs, because
+	 a given character might be encoded at several locations and kerning
+	 should still work. As a worst case estimation, we allocate 256^2
+	 and realloc later. */ 
       if ((pFontBase->pFontArray[FontID].pKernMap=
-	   (KERN_ENTRY *)malloc( k*sizeof( KERN_ENTRY)))==NULL){
-	sprintf( err_warn_msg_buf, "Error allocating memory for kerning map (FontID=%d)",
+	   (METRICS_ENTRY *)malloc( (256*256) *sizeof( METRICS_ENTRY)))==NULL){
+	sprintf( err_warn_msg_buf, "Error allocating memory for metrics map (FontID=%d)",
 		 FontID);
 	T1_PrintLog( "T1_LoadFont()", err_warn_msg_buf,
 		     T1LOG_WARNING);
@@ -669,33 +628,41 @@ int T1_LoadFont( int FontID)
       kern_tbl=pFontBase->pFontArray[FontID].pKernMap;
       pkd=pFontBase->pFontArray[FontID].pAFMData->pkd;
       j=0;
-      for ( i=0; i<k; i++){
-	if ((char1=T1_GetEncodingIndex( FontID, pkd[i].name1))==-1)
-	  /* pair is not relevant in current encoding */
-	  continue;
-	if ((char2=T1_GetEncodingIndex( FontID, pkd[i].name2))==-1)
-	  /* pair is not relevant in current encoding */
-	  continue;
-	/* Since we get here we have a relevant pair -->
-	   Put char1 in higher byte and char2 in LSB:
-	   */
-	kern_tbl[j].chars=(char1 << 8) | char2;
-	/* We only make use of horizontal kerning */
-	kern_tbl[j].hkern=pkd[i].xamt;
-	j++;
-	/* Set the kerning flag for this char1: */
-	pFontBase->pFontArray[FontID].pFlagMap[char1] |=KERN_FLAG;
-      }
-      /* Reset the remaining of the table in case there were
-	 irrelevant pairs: */
-      for ( ; j<k; j++){
-	kern_tbl[j].chars=0;
-	kern_tbl[j].hkern=0;
-      }
+      for ( i=0; i<k; i++) {
+	/* We do not check T1_GetEncodingIndices() against the return value
+	   NULL because we just loading the font in question: */
+	l=0;
+	while ((char1=(T1_GetEncodingIndices( FontID, pkd[i].name1))[l++])!=-1) {
+	  /* pair could be relevant in current encoding */
+	  m=0;
+	  while ((char2=(T1_GetEncodingIndices( FontID, pkd[i].name2))[m++])!=-1) {
+	    /* Since we get here we have a relevant pair -->
+	       Put char1 in higher byte and char2 in LSB: */
+	    kern_tbl[j].chars=(char1 << 8) | char2;
+	    /* We only make use of horizontal kerning */
+	    kern_tbl[j].hkern=pkd[i].xamt;
+	    j++;
+	  } /* while (char2) */
+	} /* while (char1) */
+      } /* for */
+      /* We are done, realloc memory: */
+      kern_tbl=(METRICS_ENTRY*) realloc( kern_tbl, j*sizeof(METRICS_ENTRY));
+      /* We now sort the kerning array with respect to char indices */
+      qsort( kern_tbl, (size_t) j, sizeof(METRICS_ENTRY),
+	     &cmp_METRICS_ENTRY );
+      /* Finally write back pointer for the case that realloc changed the
+	 pointer */
+      pFontBase->pFontArray[FontID].pKernMap=kern_tbl;
+      pFontBase->pFontArray[FontID].KernMapSize=j;
     }
     else
       pFontBase->pFontArray[FontID].pKernMap=NULL;
-  } /* End of "if (AFM-info ..)" */
+  }
+  else { /* no AFM data */
+    pFontBase->pFontArray[FontID].pKernMap=NULL;
+    pFontBase->pFontArray[FontID].pEncMap=NULL;
+  }
+  /* End of "if (AFM-info ..)" */
   
   
   /* We have just loaded a physical font into memory, thus .... */
@@ -708,14 +675,27 @@ int T1_LoadFont( int FontID)
      found. If not encoded, set space_position to -1. */
   pFontBase->pFontArray[FontID].space_position=-1;
   i=0;
-  while (i<256){
-    if (strcmp( (char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[i].data.arrayP,
-		"space")==0){
-      /* space found at position i: */
-      pFontBase->pFontArray[FontID].space_position=i;
-      break;
+  if (pFontBase->pFontArray[FontID].pFontEnc) { /* external default encoding */
+    while (i<256) {
+      if (strcmp( (char *)pFontBase->pFontArray[FontID].pFontEnc[i],
+		  "space")==0){
+	/* space found at position i: */
+	pFontBase->pFontArray[FontID].space_position=i;
+	break;
+      }
+      i++;
+    }
+  }
+  else { /* internal encoding */
+    while (i<256) {
+      if (strcmp( (char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[i].data.arrayP,
+		  "space")==0){
+	/* space found at position i: */
+	pFontBase->pFontArray[FontID].space_position=i;
+	break;
+      }
+      i++;
     }
-    i++;
   }
   
 
@@ -724,16 +704,24 @@ int T1_LoadFont( int FontID)
     pFontBase->pFontArray[FontID].pType1Data->fontInfoP[UNDERLINEPOSITION].value.data.real;
   pFontBase->pFontArray[FontID].UndrLnThick=
     pFontBase->pFontArray[FontID].pType1Data->fontInfoP[UNDERLINETHICKNESS].value.data.real;
-  /* We have to guess a value for the typographic ascender. Since this
-     dimension is not part of the Type 1 font specification, this value
-     should  explicitly be set later by the user! */
-  ascender=(float) T1_GetCharBBox( FontID, T1_GetEncodingIndex( FontID, "d")).ury;
+
+  /* We have to set the value for the typographic ascender. If possible,
+     we get it from the afm-File. But be aware this value might be undefined!
+     This value should in any acse explicitly be set later by the user! */
+  if (pFontBase->pFontArray[FontID].pAFMData!=NULL &&
+      pFontBase->pFontArray[FontID].pAFMData->gfi!=NULL) {
+    ascender=(float) pFontBase->pFontArray[FontID].pAFMData->gfi->ascender;
+  }
+  else {
+    ascender=(float) T1_GetCharBBox( FontID, T1_GetEncodingIndex( FontID, "d")).ury;
+  }
+  
   pFontBase->pFontArray[FontID].OvrLnPos=ascender
     + (float) abs( (double)pFontBase->pFontArray[FontID].UndrLnPos);
   pFontBase->pFontArray[FontID].OvrStrkPos=ascender / 2.0;
   pFontBase->pFontArray[FontID].OvrLnThick=pFontBase->pFontArray[FontID].UndrLnThick;
   pFontBase->pFontArray[FontID].OvrStrkThick=pFontBase->pFontArray[FontID].UndrLnThick;
-  
+
   
   /* Finally, set the font size dependencies pointer to NULL since we can
      assume, that at load time of a font, no size specific data of this
@@ -763,118 +751,120 @@ int T1_LoadFont( int FontID)
 }
   
 
-/* openFontMetricsFile( FontID): Gets the fontfilename corresponding to
-   FontID, opens the corresponding afm-file and fills the data structures.
-   return-value is the value returned by the parseFile() function: */
-
-int openFontMetricsFile( int FontID)
+/* openFontMetricsFile( FontID, open_sloppy): Gets the fontfilename
+   corresponding to FontID, opens the corresponding afm-file and fills
+   the data structures. return-value is the value returned by the
+   T1lib_parseFile() function. If open_sloppy is set, the minimum
+   information needed is read from AFM file. This can be considered a
+   fallback for problematic AFM files. */
+static int openFontMetricsFile( int FontID, int open_sloppy)
 {
   char *FontFileName;
   char *AFMFileName;
+  char *afm_name;
   char *AFMFileNamePath;
   
-  int i;
+  int i, j;
   FILE *metricsfile;
   
 
-  FontFileName=T1_GetFontFileName( FontID);
-  i=strlen(FontFileName);
-  AFMFileName=(char *)malloc( i+1);
-  strcpy( AFMFileName, FontFileName);
-  /* We now change the extension pfa or pfb to afm: */
-  AFMFileName[i-3]='a';
-  AFMFileName[i-2]='f';
-  AFMFileName[i-1]='m';
-
-  /* Get full path of the afm file */
-  AFMFileNamePath=Env_GetCompletePath( AFMFileName, T1_AFM_ptr);
+  afm_name=T1_GetAfmFileName(FontID);
+  
+  if (afm_name!=NULL) { /* We have name explicitly specified */
+    /* It needs to be freeable */
+    if ((AFMFileName=
+	 (char *)malloc( (strlen(afm_name)+1)*sizeof( char)))==NULL) {
+      T1_errno=T1ERR_ALLOC_MEM;
+      return( -6);
+    }
+    strcpy( AFMFileName, afm_name);
+  }
+  else {
+    FontFileName=T1_GetFontFileName( FontID);
+    i=strlen(FontFileName);
+    j=i;
+    AFMFileName=(char *)malloc( i+5);
+    strcpy( AFMFileName, FontFileName);
+    while ( AFMFileName[i] != '.'){
+      if (i==0) break;
+      else i--;
+    }
+    if (i==0){
+      /* We have a filename without extension -> append extension */
+      AFMFileName[j]='.';
+      AFMFileName[j+1]='a';
+      AFMFileName[j+2]='f';
+      AFMFileName[j+3]='m';
+      AFMFileName[j+4]='\0';
+    }
+    else{
+      /* we found a '.' -> replace extension */
+      AFMFileName[i+1]='a';
+      AFMFileName[i+2]='f';
+      AFMFileName[i+3]='m';
+      AFMFileName[i+4]='\0';
+    }
+  }
+  
+  /* Get full path of the afm file (The case of a full path name
+     name specification is valid */
+  AFMFileNamePath=intT1_Env_GetCompletePath( AFMFileName, T1_AFM_ptr);
   free( AFMFileName);
   
   /* open afm-file: */
-  if ((metricsfile=fopen(AFMFileNamePath,"r"))==NULL){
-    if (AFMFileNamePath!=NULL)
+  if (AFMFileNamePath!=NULL){
+    if ((metricsfile=fopen(AFMFileNamePath,"rb"))==NULL){
       free(AFMFileNamePath);
-    return(-4);
+      return(-4);
+    }
+    else {
+      free(AFMFileNamePath);
+    }
+  }
+  else{
+    return( -5);
   }
-  free(AFMFileNamePath);
   
   /* Call procedure to read afm-file and store the data formatted.
      Flags used here: P_M  All Metrics Information
                       P_P  Pair Kerning Information
+		      P_C  Composite Character Data (since t1lib V.1.2)
      The P_G flag to get global font information should not be used
      if not absolutely needed. When parsing an unknown keyword, which
-     may be harmless, the parseFile function returns the error code
+     may be harmless, the T1lib_parseFile function returns the error code
      -1 (parseError). On the other hand, all other really relevant
      data may habe been parsed and stored correctly. In such a case,
      There's no way to make a serious decision whether an error has
      occured or not.
      */
-  i=parseFile( (FILE *) metricsfile,
-	       (FontInfo **) &(FontBase.pFontArray[FontID].pAFMData),
-	       P_M | P_P );
+  if (open_sloppy!=0)
+    i=T1lib_parseFile( (FILE *) metricsfile,
+		       (FontInfo **) &(FontBase.pFontArray[FontID].pAFMData),
+		       P_M );
+  else
+    i=T1lib_parseFile( (FILE *) metricsfile,
+		       (FontInfo **) &(FontBase.pFontArray[FontID].pAFMData),
+		       P_G | P_M | P_P | P_C );
   fclose(metricsfile);
   return(i);
 }
 
 
 
-/* openFontMetricsFileSloppy( FontID): Gets the fontfilename corresponding to
-   FontID, opens the corresponding afm-file and tries to fill the data
-   structures only with metric data. This can be considered a fallback!
-   */
-int openFontMetricsFileSloppy( int FontID)
-{
-  char *FontFileName;
-  char *AFMFileName;
-  char *AFMFileNamePath;
-  int i;
-  FILE *metricsfile;
-  
-
-  FontFileName=T1_GetFontFileName( FontID);
-  i=strlen(FontFileName);
-  AFMFileName=(char *)malloc( i+1);
-  strcpy( AFMFileName, FontFileName);
-  /* We now change the extension pfa or pfb to afm: */
-  AFMFileName[i-3]='a';
-  AFMFileName[i-2]='f';
-  AFMFileName[i-1]='m';
-
-  /* Get full path of the afm file */
-  AFMFileNamePath=Env_GetCompletePath( AFMFileName, T1_AFM_ptr);
-  free( AFMFileName);
-  
-  /* open afm-file: */
-  if ((metricsfile=fopen(AFMFileNamePath,"r"))==NULL){
-    /* patch #1 */
-    if (AFMFileNamePath!=NULL)
-      free(AFMFileNamePath);
-    return(-4);
-  }
-  free(AFMFileNamePath);
-  
-  /* Call procedure to read afm-file and store the data formatted.
-     Flags used here: P_M  All Metrics Information
-     */
-  i=parseFile( (FILE *) metricsfile,
-	       (FontInfo **) &(FontBase.pFontArray[FontID].pAFMData),
-	       P_M );
-  fclose(metricsfile);
-  return(i);
-}
-
-
-
-
-/* CreateNewFontSize( FontID, size): Create a new size "size" of font
+/* T1int_CreateNewFontSize( FontID, size): Create a new size "size" of font
    "FontID" and allocate all data necessary for this. The data
    structure is connected to the linked list of FontSizeDeps for this
    font. Returns a pointer to the newly created FontSizeDeps-struct
    if all went correct and NULL otherwise.
    Since of version 0.3 a member antialias has been added to the
-   FONTSIZEDEPS structure!
+   FONTSIZEDEPS structure! This can be:
+
+   0:     bitmaps are stored in this struct
+   1:     non-antialiased bytemaps are stored in this struct
+   2:     low-antialiased bytemaps are stored in this struct
+   4:     high-antialiased bytemaps are stored in this struct
    */
-FONTSIZEDEPS *CreateNewFontSize( int FontID, float size, int aa)
+FONTSIZEDEPS *T1int_CreateNewFontSize( int FontID, float size, int aa)
 {
 
   FONTSIZEDEPS *pFontSizeDeps, *pPrev;
@@ -883,7 +873,7 @@ FONTSIZEDEPS *CreateNewFontSize( int FontID, float size, int aa)
   /* First, get to the last font size in the linked list for this font.
      The following routine returns the address of the last struct in the
      linked list of FONTSIZEDEPS or NULL if none exists. */
-  pFontSizeDeps=GetLastFontSize( FontID);
+  pFontSizeDeps=T1int_GetLastFontSize( FontID);
   pPrev=pFontSizeDeps;
   
   
@@ -909,10 +899,7 @@ FONTSIZEDEPS *CreateNewFontSize( int FontID, float size, int aa)
   /* Put the size into this structure */
   pFontSizeDeps->size=size;
   /* Set the antialias mark: */
-  if (aa)
-    pFontSizeDeps->antialias=1;
-  else
-    pFontSizeDeps->antialias=0;
+  pFontSizeDeps->antialias=aa;
   
   /* Just the current becomes now the last item in the linked list: */
   pFontSizeDeps->pNextFontSizeDeps=NULL;
@@ -932,11 +919,6 @@ FONTSIZEDEPS *CreateNewFontSize( int FontID, float size, int aa)
 	      pFontBase->pFontArray[FontID].FontTransform[1],
 	      pFontBase->pFontArray[FontID].FontTransform[2],
 	      pFontBase->pFontArray[FontID].FontTransform[3]);
-  /* Apply device specific scaling */
-  pFontSizeDeps->pCharSpaceLocal=(struct XYspace *)
-    Scale(pFontSizeDeps->pCharSpaceLocal,
-	  DeviceSpecifics.scale_x,
-	  - DeviceSpecifics.scale_y);
   /* Apply desired scaling factor, and make it Permanent */
   pFontSizeDeps->pCharSpaceLocal=(struct XYspace *) Permanent
     (Scale(pFontSizeDeps->pCharSpaceLocal, size, size));
@@ -958,10 +940,10 @@ FONTSIZEDEPS *CreateNewFontSize( int FontID, float size, int aa)
   
 
 
-/* QueryFontSize( FontID, size, aa): Search if a requested size of font
+/* T1_QueryFontSize( FontID, size, aa): Search if a requested size of font
    FontID is already existing. If so, it returns a pointer to the
    respective FontSizeDeps-structure,  otherwise NULL is returned: */
-FONTSIZEDEPS *QueryFontSize( int FontID, float size, int aa)
+FONTSIZEDEPS *T1int_QueryFontSize( int FontID, float size, int aa)
 {
   
   FONTSIZEDEPS *link_ptr;
@@ -986,10 +968,10 @@ FONTSIZEDEPS *QueryFontSize( int FontID, float size, int aa)
   
 }
 
-/* FONTSIZEDEPS *GetLastFontSize( FontID): Get the address of the
+/* FONTSIZEDEPS *T1int_GetLastFontSize( FontID): Get the address of the
    last struct in the linked list of FontSizeDeps or NULL if there is
    no existing size dependent data. */
-FONTSIZEDEPS *GetLastFontSize( int FontID)
+FONTSIZEDEPS *T1int_GetLastFontSize( int FontID)
 {
   FONTSIZEDEPS *link_ptr, *result_ptr;
   
@@ -1010,3 +992,17 @@ FONTSIZEDEPS *GetLastFontSize( int FontID)
   return((FONTSIZEDEPS *)(result_ptr));
 }
 
+
+/* A function for comparing METRICS_ENTRY structs */
+static int cmp_METRICS_ENTRY( const void *entry1, const void *entry2)
+{
+  if (((METRICS_ENTRY *)entry1)->chars <
+      ((METRICS_ENTRY *)entry2)->chars)
+    return(-1);
+  if (((METRICS_ENTRY *)entry1)->chars >
+      ((METRICS_ENTRY *)entry2)->chars)
+    return(1);
+  return(0); /* This should not happen */
+}
+    
+    
